#!/usr/bin/perl 

use XML::Parser;
use Data::Dumper;
use POSIX;
use PDF;
use strict;

my $GSDIR="/usr/local/share/ghostscript/fonts";
my $HelveticaNarrow="NimbusSanL-ReguCond";

my %Picture_Defs = (
		    'leftMargin' => 12, # 12 points
		    'rightMargin' => 6, # 6 points
		    'topMargin' => 12, # 12 points
		    'width' => 12 * 27, # 27 pica
		    'headSize' => 12, # 12 points
		    'bodySize' => 8,  # 8 points
		    'headFont' => "Helvetica-Bold",
		    'bodyFont'     => "Helvetica",
		    'legendFont'  => "$HelveticaNarrow",
		    'lineWidth' => 7, # 7 points
		    'lineSkip'  => 3, # 3 points between lines
		    'paragraphSkip' => 6, # 6 points between paragraphs
		    'legendLength' => 15, # 15 points long lines in legend
		    'legendOffset' => 60, # Offset between two legend parts
		    'legendStart' => 36, # Start of first legend line
		    'legendMargin' => 6, # distance between two legends
		 );

my (@benchmarks, %bench, $tsub, $tsuper, $flag);
my $p1 = new XML::Parser(Style => 'Stream');

$p1->parsefile('benchmarks.xml');


# print Dumper(@benchmarks);

do_bonnie( \@benchmarks );

sub StartTag {
  my $p = shift;
  my $t = shift;
  my %atts = %_;
  my $k;
#  print "Start: $t, flag: $flag>\n";
  if ($t eq "bench") {
    foreach $k (keys %atts) {
      $bench{"machine"}{$k} = $atts{$k}
    }
  } elsif (($t eq "nonr") || ($t eq "nomr") || 
	   ($t eq "monr") || ($t eq "momr")) {
    $tsub = $t;
    $tsuper = $p->current_element;
  } elsif (($t eq "sys") || ($t eq "user") ||
	   ($t eq "real") || ($t =~ /[bc](read|write)/) ||
	   ($t eq "seeks")) {
    $tsuper = $p->current_element;
    $tsub = $t;
  } elsif ($t eq "bonnie") {
    foreach $k (keys %atts) {
      $bench{$t}{$k} = $atts{$k};
    }
  }
}

sub EndTag {
  my $p = shift;
  my $t = shift;
#  print "End: $t, flag: <$flag>\n";
  if (($t eq "nonr") || ($t eq "nomr") || 
	   ($t eq "monr") || ($t eq "momr")) {
    undef $tsub;
  } elsif ($t eq "bench") {
    push @benchmarks, { %bench };
    undef %bench;
  } elsif (($t eq "sys") || ($t eq "user") ||
	   ($t eq "real") || ($t =~ /[bc](read|write)/) ||
	   ($t eq "seeks")) {
    undef $tsub;
  } elsif (($t eq "utah") || ($t eq "testc") ||
	   ($t eq "mpbyte") || ($t eq "tools") ||
	   ($t eq "mem") || ($t eq "call") ||
	   ($t eq "pipe") || ($t eq "fexec") ||
	   ($t eq "bonnie")) {
    undef $tsuper;
  }
}


sub Text {
  my $parser = shift;
  my $chardata = $_;
#  print "Text: $c, tsuper: <$tsuper>, tsub: <$tsub>\n";
  $bench{$flag} = $chardata if (defined $flag);
  $bench{$tsuper}{$tsub} = $chardata if (defined $tsub);
}


sub do_bonnie {

  #
  # Hash fr Legendenbeschreibung:
  # erster Teil des Werts ist die Farbe als Array-Referenz
  # zweiter Teil ist die linke Legendenbeschriftung
  # dritter Teil ist die rechte Legendenbeschriftung
  #
  my %legend = ( 'cwrite' => [[0.5, 0.5, 1.0],
			      "Schreiben","zeichenweise"], # light blue
		 'bwrite' => [[0.0, 0.0, 1.0], 
			      "(MByte/S)", "blockweise"], # dark blue
		 'cread'  => [[1.0, 0.5, 0.0], 
			      "Lesen", "zeichenweise"],# orange
		 'bread'  => [[1.0, 0.0, 0.0], 
			      "(MByte/s)", "blockweise"],# dark red
		 'seeks'  => [[1.0, 1.0, 0.0], 
			      "", "Zugriffe pro Sekunde"],# yellow
	       );

  #
  # Zahlenformat auf Deutsch setzen, damit Perl "," statt "." 
  # in den Zahlen versteht.
  #

  POSIX::setlocale( &POSIX::LC_NUMERIC, "de" );

  my ($benches) = @_;
  my (@bonnie_results, $x, $t, $b, $maxMB, $maxSeek, $length);
  my $benchCount = scalar(@$benches);

  #
  # Bonnie-Teile aus den Benchmarks in Array packen
  #
  for (my $i = 0; $i < $benchCount; $i++) {
    push @bonnie_results , 
    exists(($benches->[$i])->{"bonnie"})? ($benches->[$i])->{"bonnie"} : undef; 
  }

  #
  # Maximum der jeweiligen Teilwerte bestimmen
  # Undefinierte Eintrge (fehlender Bonnie-Wert!)
  # berspringen, damit kein leerer Hash erzeugt wird
  #
  foreach $t ("cwrite", "bwrite", "cread", "bread") {
    $maxMB = -1.0;
    foreach $b (@bonnie_results) {
      $maxMB = $b->{$t} if $b && $maxMB < $b->{$t};
    }
  }
  $maxSeek = -1.0;
  foreach $b (@bonnie_results) {
    $maxSeek = $b->{'seeks'} if $b && $maxSeek < $b->{'seeks'};
  }

  #
  # Zahlenformat wieder auf ANSI stellen, damit korrektes
  # PostScript entsteht 
  #
  POSIX::setlocale( &POSIX::LC_NUMERIC, "C" );

  my $totalwidth = 0.90 * ($Picture_Defs{'width'} - 
			   $Picture_Defs{'leftMargin'} -
			   $Picture_Defs{'rightMargin'});

  #
  # PDF-Objekt erzeugen und initialisieren
  #
  my $pdf = new PDF;
  die "Konnte PDF-Datei nicht erzeugen" 
    if ($pdf->open_file("bonnie.pdf") == -1);
  $pdf->set_info("Creator",$0);
  $pdf->set_info("Author","Christian Kirsch");
  $pdf->set_info("Title","Bonnie-Ergebnisse");
  $pdf->set_parameter("FontAFM","$HelveticaNarrow=$GSDIR\/n019043l.afm");
  $pdf->set_parameter("FontOutline","$HelveticaNarrow=$GSDIR\/n019043l.pfa");
  
  $pdf->begin_page("A4");
  #
  # Startpunkt der ersten Zeile berechnen
  #
  my $pageh = @{$pdf->get_pagesize()}[1];

  $x = $Picture_Defs{'leftMargin'};

  my $image = $pdf->open_GIF("iXwhite_min.gif");
  my $imageh = @{$pdf->get_image_size($image)}[1];

  my $top; 
  my $minSkip = $Picture_Defs{'topMargin'} +   $Picture_Defs{'headSize'} * 2;
  if ($imageh > $minSkip) {
    $top = $pageh - $imageh - $Picture_Defs{'topMargin'};
  } else {
    $top = $pageh - $minSkip;
  }
  $pdf->place_image($image,$x,$top);
  $pdf->set_font($Picture_Defs{'headFont'}, 
		$Picture_Defs{'headSize'}*1.4);

  $x += @{$pdf->get_image_size($image)}[0] + 3;

  $pdf->text("Bonnie-Ergebnisse",$x,$top);
  $top -= $Picture_Defs{'lineSkip'} + $Picture_Defs{'headSize'}*1.4;
  # 
  # Schleife ber alle Benchmarks
  #
  for (my $i = 0; $i < $benchCount; $i++) {
    #
    # Benchmarks mit fehlendem Bonnie berspringen
    #
    next unless $bonnie_results[$i];

    $x = $Picture_Defs{'leftMargin'};
    #
    # berschrift erzeugen
    # <Maschine>, <MHz>, <CPU>, <OS>
    #
    my $mach = $benches->[$i]->{"machine"};
    my $head = "$mach->{'machine'}";
    $pdf->set_font($Picture_Defs{'headFont'}, 
		  $Picture_Defs{'headSize'});
    $pdf->text($head, $x, $top);
    $pdf->set_font( $Picture_Defs{'bodyFont'}, 
		   $Picture_Defs{'headSize'});
    $pdf->text(", $mach->{'mhz'} MHz $mach->{'cputype'} ".
	       "$mach->{'ram'} MByte".
	      " ($mach->{'os'})");
    $top -= ($Picture_Defs{'lineSkip'}+$Picture_Defs{'headSize'});
    $x = $Picture_Defs{'leftMargin'};
    $pdf->set_font($Picture_Defs{'bodyFont'}, $Picture_Defs{'bodySize'});

    #
    # Jetzt die Teilergebnisse dieses Benchmarks ausgeben
    #
    foreach $t ("cwrite", "bwrite", "cread", "bread", "seeks") {
      my $max = ($t eq "seeks"  ? $maxSeek : $maxMB );
      my $length = $totalwidth * ( $bonnie_results[$i]->{$t} / $max );
      my $c = $legend{$t}[0];
      $pdf->set_rgbcolor_fill($c);
      $pdf->fill_rect($x, $top, $length, $Picture_Defs{'lineWidth'});
      $pdf->set_rgbcolor_fill('black');
      $pdf->set_text_pos($x+$length+$Picture_Defs{'leftMargin'}, $top);
      $pdf->text($bonnie_results[$i]->{$t});
      $top -= ($Picture_Defs{'lineWidth'} + $Picture_Defs{'lineSkip'});
    }
    $top -= ($Picture_Defs{'headSize'} + $Picture_Defs{'paragraphSkip'});

  } # for $i ... $benchCount
  #
  # Trennlinie zwischen Ergebnissen und Legende
  $top += $Picture_Defs{'paragraphSkip'};
  $pdf->set_linewidth(1.0);
  $pdf->set_rgbcolor_stroke('black');
  $pdf->line($x, $top,$x+$totalwidth, $top);
  $top -= ($Picture_Defs{'paragraphSkip'} + $Picture_Defs{'bodySize'});
  $pdf->set_linewidth($Picture_Defs{'lineWidth'});
  $length = $Picture_Defs{'legendLength'};
  $pdf->set_font($Picture_Defs{'legendFont'}, 
		$Picture_Defs{'bodySize'});

  my @tpos = ( $top, 
	       $top - $Picture_Defs{'bodySize'} - 
	       $Picture_Defs{'lineSkip'} );
  #
  # Legende ausgeben
  #
  # Dazu bestimmen wir erst die grte Breite der linken und
  # rechten auszugebenden Texte. Der "Seeks"-Teil interessiert
  # nicht, da dessen Beschriftung nicht relativ zu anderen auszurichten ist
  #
  my ($leftmax,$rightmax) = (-1, -1);
  my ($left, $right, $w1, $w2);
  foreach $t ("cwrite", "bwrite", "cread", "bread") {
    
    $left = $legend{$t}[1];
    $w1 = textWidth($pdf,$left);
    $leftmax = $w1 if $w1 > $leftmax;

    $right = $legend{$t}[2];
    $w2 = textWidth($pdf,$right);
    $rightmax = $w2 if $w2 > $rightmax;

  }

  $x = $Picture_Defs{'leftMargin'};
  my $ind = 0;
  my $bs = $Picture_Defs{'bodySize'} / 2;
  foreach $t ("cwrite", "bwrite", "cread", "bread", "seeks") {
    my $x1 = $x  + $leftmax + $bs;
    my $x2 = $x1 + $length  + $bs;

    $left = $legend{$t}[1];
    $right = $legend{$t}[2];

    $top = $tpos[$ind];
    $pdf->set_rgbcolor_fill('black');
    $pdf->text($left,$x,$top);
    $pdf->set_rgbcolor_fill($legend{$t}[0]);
    $pdf->rect( $x1,$top, $length, 
	       $Picture_Defs{'lineWidth'});
    $pdf->fill();
    $pdf->set_rgbcolor_fill('black');
    $pdf->text($right, $x2 , $top);
    $ind = 1- $ind;
    if ($ind == 0) {
      $x +=  $length  + $rightmax + 4*$bs;
      $x += $leftmax + $bs unless $t eq "bread";
    }
  }
  
  $pdf->end_page();
#  $pdf->closefile();
#  $pdf->deleteobj();
#  PDF_close($pdf);
}

sub textWidth {
  my ($pdf,$text) = @_;
#  my $font = PDF_get_font($pdf);
 # my $size = PDF_get_fontsize($pdf);
  return $pdf->get_textwidth($text);
}

sub textLeft {
  my ($pdf, $x, $y, $text) = @_;
  PDF_set_text_pos($pdf,$x,$y);
  PDF_show($pdf,$text);
}
